# -*- coding: utf-8 -*-
"""
Tutorial illustrating the use of the GRAPL library, for representing, analyzing and processing
acyclic directed mixed graphs for structural causal modelling.

(CC BY-SA 4.0) 2021. If you use this code, please cite: M.A. Little, R. Badawy,
2019, "Causal bootstrapping", arXiv:1910.09648
"""

import grapl.algorithms as algs
import grapl.dsl as dsl

# Create a GRAPL DSL parser
grapl_obj = dsl.GraplDSL()

# Create a DAG from a GRAPL string
dag_grapl = ' "Simple back-door graph"; \
    C; X; Y; \
    C -> X; \
    C -> Y; \
    X -> Y; '
G = grapl_obj.readgrapl(dag_grapl)
G.display()

print('\nCheck if the DAG is acyclic:')
print(G.isdag()) # True

print('\nFactorized joint distribution:')
fac_str, expr, isdag = algs.dagfactor(G, simplify=False)
print(fac_str) # p(X,C,Y)=[p(Y|X,C)p(X|C)p(C)]

print('\nInterventional (cause-effect) distribution of X on Y:')
id_str, expr, isdag = algs.truncfactor(G, {'X'}, {'Y'})
print(id_str) # p_{X}(Y)=\sum_{C}[p(Y|X,C)p(C)]



# -*- coding: utf-8 -*-
"""
A small library of functions for representing, analyzing and processing
acyclic directed mixed graphs for structural causal modelling.

(CC BY-SA 4.0) 2021. If you use this code, please cite: M.A. Little, R. Badawy,
2019, "Causal bootstrapping", arXiv:1910.09648
"""

from queue import Empty
import grapl.util as util

class ADMGNode():
    """An ADMG node (random variable) object. The member variables, parents, children,
       bidirects are Sets of Strings, each string is the name of a random variable (node)
       in the graph. Bidirects are (nodes) connected to this one via a latent edge. The
       properties member is an optional Set of properties.
    """
    def __init__(self, parents={}, bidirects={}, children={}, properties={}):
        self.parents = set(parents)
        self.children = set(children)
        self.bidirects = set(bidirects)
        self.properties = set(properties)

class ADMG():
    """An ADMG graph object. The title member variable (String) is the name of graph
       (mostly for display purposes). The vars member is a Dict of String-ADMGNode pairs,
       representing his graph's random variables (ADMGNodes), indexed by node name (String).
    """
    def __init__(self, title=''):
        self.title = title
        self.vars = dict()

    def settitle(self, title=''):
        """Set graph title (String)."""
        self.title = title

    def addvar(self, name='', parents={}, bidirects={}, children={}, properties={}):
        """Add a new random variable to the graph with string name, with optional parents, bidirects
           children and/or properties. The name must not be empty.
           Each random variable (parents, bidirects, children) is a Set of Strings,
           each string representing a random variable in the graph.
        """
        if name != '':
            self.vars[name] = ADMGNode(parents,bidirects,children,properties)
        else:
            raise ValueError('admg:addvar: Invalid node name')

    def addedges(self, name='', parents={}, bidirects={}, children={}):
        """Add (optional) new edges connecting parents, bidirects and/or child random variables,
           to the random variable given in name (which must not be empty).
           Edges are Sets of Strings, each string representing a random variable in the graph.
        """
        if name != '':
            self.vars[name].parents = self.vars[name].parents.union(parents)
            self.vars[name].children = self.vars[name].children.union(children)
            self.vars[name].bidirects = self.vars[name].bidirects.union(bidirects)
        else:
            raise ValueError('admg:addedges: Invalid node name')

    def connect(self):
        """Ensure all parent/child/bidirect edges are consistent across the graph.
           Explicitly: every parent of a random variable, must have that variable
           as a child, and every random variable connected by a latent (bidrected)
           edge to another random variable, must have a bidirected edge to that
           first random variable.
        """
        for node in set(self.vars):
            for p in self.pa({node}):
                self.vars[p].children = self.ch({p}).union({node})
            for b in self.bi({node}):
                self.vars[b].bidirects = self.bi({b}).union({node})

    def display(self):
        """Pretty-print a text representation of the graph (ADMG) object."""
        nodes = set(self.vars)
        print('Title: ' + self.title)
        print('Vars: ' + util.csepstr(nodes))
        print('Parents:')
        for node in nodes:
            parents = self.pa({node})
            if len(parents) > 0:
                print(node,'<-', util.csepstr(parents))
        print('Children:')
        for node in nodes:
            children = self.ch({node})
            if len(children) > 0:
                print(node,'->', util.csepstr(children))
        print('Bidirects:')
        for node in nodes:
            bidirects = self.bi({node})
            if len(bidirects) > 0:
                print(node,'<-->', util.csepstr(bidirects))

    def nodes(self):
        """Return set of all random variables (nodes) in the graph.
           This is represented as a Set of Strings, each string being the name of
           a random variable.
        """
        return set(self.vars)

    def sub(self, subvars):
        """Find a subgraph of a graph. The subvars parameter is a Set of Strings,
           each string naming a random variable. If there are no such random variables
           in the graph, the result will be an empty graph. Returns an ADMG object.
        """
        sub = ADMG(self.title)
        for node in set(self.vars):
            if node in subvars:
                parents = self.pa({node}).intersection(subvars)
                bidirects = self.bi({node}).intersection(subvars)
                sub.vars[node] = ADMGNode(parents,bidirects,self.vars[node].properties)
        sub.connect()
        return sub
    
    def ch(self, nodes):
        """Find the union of all children of a set of random variables (nodes).
           The set of nodes is given as a Set of Strings, each string being the
           name of a random variable. Each such random variable must exist in the graph.
           Returns a Set of Strings, each string being the name of a random variable.
        """
        try:
            children = set()
            for node in nodes:
                children = children.union(self.vars[node].children)
            return children
        except KeyError as err:
            raise ValueError(f"admg.ch:Var '{node}' does not exist")

    def pa(self, nodes):
        """Find the union of all parents of a set of random variables (nodes).
           The set of nodes is given as a Set of Strings, each string being the
           name of a random variable. Each such random variable must exist in the graph.
           Returns a Set of Strings, each string being the name of a random variable.
        """
        try:
            parents = set()
            for node in nodes:
                parents = parents.union(self.vars[node].parents)
            return parents
        except KeyError as err:
            raise ValueError(f"admg.pa:Var '{node}' does not exist")

    def bi(self, nodes):
        """Find the union of all bidirected (latent) edges of a set of random variables (nodes).
           The set of nodes is given as a Set of Strings, each string being the
           name of a random variable. Each such random variable must exist in the graph.
           Returns a Set of Strings, each string being the name of a random variable.
        """
        try:
            bidirects = set()
            for node in nodes:
                bidirects = bidirects.union(self.vars[node].bidirects)
            return bidirects
        except KeyError as err:
            raise ValueError(f"admg.bi:Var '{node}' does not exist")
    
    def an(self, nodes):
        """Find the union of all ancestors of a given set of random variables (nodes),
           and the nodes themselves. The set of nodes is given as a Set of Strings,
           each string being the name of a random variable. Each such random variable must
           exist in the graph. Returns a Set of Strings, each string being the name
           of a random variable.
        """

        def _an_recurse(node, ancestors):
            parents = self.pa({node})
            ancestors = ancestors.union(parents)
            for parent in parents:
                ancestors = _an_recurse(parent, ancestors)
            return ancestors

        ancestors = set()
        for node in nodes:
            ancestors = ancestors.union(_an_recurse(node, set()))
        return ancestors.union(nodes)

    def de(self, nodes):
        """Find the union of all descendants of a given set of random variables (nodes),
           and the nodes themselves. The set of nodes is given as a Set of Strings,
           each string being the name of a random variable. Each such random variable must
           exist in the graph. Returns a Set of Strings, each string being the name
           of a random variable.
        """

        def _de_recurse(node, descendants):
            children = self.ch({node})
            descendants = descendants.union(children)
            for child in children:
                descendants = _de_recurse(child, descendants)
            return descendants

        descendants = set()
        for node in nodes:
            descendants = descendants.union(_de_recurse(node, set()))
        return descendants.union(nodes)
    
    def nd(self, nodes):
        """Find all non-descendants of a given set of random variables (nodes). The set of nodes
           is given as a Set of Strings, each string being the name of a random variable.
           Each such random variable must exist in the graph. Returns a Set of Strings, each string
           being the name of a random variable.
        """
        return self.nodes().difference(self.de(nodes))

    def dispa(self, node):
        """Find the district parents of a random variable (node) in a graph. The node is given
           as a String, which is the name of the random variable. This random variable must
           exist in the graph. Returns a Set of Strings, each string being the name of a random
           variable.
        """
        district = self.district(node)
        parents = self.pa(district)
        return parents.union(district.difference({node}))

    def mb(self, node):
        """Find the Markov blanket of a random variable (node) in a DAG. The node is given
           as a String, which is the name of the random variable, which must exist in the graph.
           If the graph is a DAG, returns a Set of Strings, each string being the name of a random
           variable, otherwise returns None.
        """
        if not self.isdag():
            return None
        else:
            children = self.ch({node})
            mb_nodes = self.pa({node}).union(children)
            for child in children:
                mb_nodes = mb_nodes.union(self.pa({child}))
            return mb_nodes.difference({node})

    def districts(self):
        """Compute all districts (c-components) of an ADMG. Returns a List of Set of Strings,
           each Set being a district of the graph. The sets are names of the random variables
           in each district.
        """
        components = []
        nodes = set(self.vars)
        while len(nodes) > 0:
            next_node = nodes.pop()
            component = self.district(next_node)
            components.append(component)
            nodes = nodes.difference(component)
        return components
            
    def district(self, node):
        """Find the district (c-components) of a random variable (node). The node is given
           as a String, which is the name of the random variable, which must exist in the graph.
           Returns a Set of Strings, each string being the name of a random variable in the
           district.
        """
        
        def _district_recurse(node, component):
            component = component.union({node})
            bidirects = self.bi({node})
            for bidirect in bidirects:
                if bidirect not in component:
                    component = _district_recurse(bidirect, component)
            return component
        
        return _district_recurse(node, set())

    def fix(self, node):
        """Apply the interventional fixing operation to a random variable (node) in the
           graph. This changes the graph (ADMG) object. The node is given as the string
           name of the random variable, which must exist in the graph. 
        """
        for parent in self.pa({node}):
            self.vars[parent].children = self.vars[parent].children.difference({node})
        self.vars[node].parents = set()
        for bidirect in self.bi({node}):
            self.vars[bidirect].bidirects = self.vars[bidirect].bidirects.difference({node})
        self.vars[node].bidirects = set()

    def isfixable(self, node):
        """Check whether it is possible to apply the interventional fixing operation
           to a random variable (node). The node is given as the string
           name of the random variable, which must exist in the graph. Returns True
           if fixable, False if not.
        """
        district = self.district(node)
        descend = self.de({node}).union({node})
        fix = district.intersection(descend)
        return (fix == {node})

    def isdag(self):
        """Returns True if graph is a DAG (no bidirects => no latent variables)."""
        bidirects = False
        for node in set(self.vars):
            bidirects = bidirects or (len(self.bi({node})) > 0)
        return (not bidirects)
    
    def fixable(self):
        """Returns set of all fixable nodes in the graph. Returns a Set of Strings,
           each string being the name of a fixable random variable in the graph.
        """
        nodes = set(self.vars)
        fix_set = set()
        for node in nodes:
            if self.isfixable(node):
                fix_set = fix_set.union({node})
        return fix_set

    def topsort(self):
        """List of random variables (nodes) sorted topologically. Returns a List of Strings,
           each string being the name of a random variable.
        """

        # Recursive depth-first search
        def _visit(node):
            if node not in visited:
                children = self.ch({node})
                for child in children:
                    _visit(child)
                visited.add(node)
                sorted.insert(0, node)

        visited = set()
        sorted = list()
        nodes = set(self.vars)

        # Initiate depth-first search from all possible starting points
        while True:
            unvisited = nodes.difference(visited)
            if unvisited:
                _visit(unvisited.pop())
            else:
                break

        return sorted

    def isayclic(self):
        """Returns True if the graph is acyclic. Attempt to sort nodes topologically,
           and uses this search to detect if graph is cyclic."""

        # Recursive depth-first search, marking nodes which are part of a cyclic path
        def _visit(node):
            if node in visited:
                return False
            if node in marked:
                return True
            marked.add(node)
            cycle = False
            children = self.ch({node})
            for child in children:
                cycle = _visit(child)
                if cycle:
                    break
            if cycle:
                return True
            marked.discard(node)
            visited.add(node)
            return False

        visited = set()
        marked = set()
        nodes = set(self.vars)

        # Initiate depth-first search from all possible starting points, with early termination
        # if cycle detected
        cyclic = False
        while not cyclic:
            unvisited = nodes.difference(visited)
            if unvisited:
                cyclic = _visit(unvisited.pop())
            else:
                break

        return not cyclic



# -*- coding: utf-8 -*-
"""
Tutorial illustrating the use of the GRAPL library, for representing, analyzing and processing
acyclic directed mixed graphs for structural causal modelling.

(CC BY-SA 4.0) 2021. If you use this code, please cite: M.A. Little, R. Badawy,
2019, "Causal bootstrapping", arXiv:1910.09648
"""

import grapl.algorithms as algs
import grapl.dsl as dsl

# Create a GRAPL DSL parser
grapl_obj = dsl.GraplDSL()

# Load an ADMG description from a GRAPL file
G = grapl_obj.readgrapl(open('grapl/graphs/richardson_2017.grapl', 'r').read())
G.display()

print('\nCheck if the ADMG is acyclic:')
print(G.isayclic()) # True

print('\nTopological sort of observed nodes:')
print(G.topsort()) # ['X1', 'X2', 'X3', 'X4']

print('\nCompute all ADMG districts:')
print(G.districts()) # [{'X3'}, {'X2', 'X4'}, {'X1'}]

print('\nInterventional (cause-effect) distribution of X2 on X4:')
id_str, expr, isident = algs.idfixing(G, {'X2'}, {'X4'})
if isident:
    print(id_str) # p_{X2}(X4)=\sum_{X3,X2',X1',X1}[p(X4|X3,X2',X1')p(X3|X2,X1)p(X2',X1')p(X1)]
else:
    print('Interventional distribution not identifiable')



# -*- coding: utf-8 -*-
"""
Implements a lexer and parser of a simple domain-specific language
(GRAPL -- GRAph Processing Language) for describing acyclic directed mixed graphs.
A quasi-EBNF description of the language goes as follows, where 'grapl' is the top
level rule:

        grapl = TITLE EOC graph | graph
        graph = graph command | command
        command = NODE EOC | NODE DIR_EDGE NODE EOC | NODE BI_EDGE NODE EOC

The terminal identifiers are:

        TITLE: any string of characters enclosed in quotes
        NODE: any non-empty string of alphabetical characters including underline,
              followed optionally by any natural number
        EOC: ';', the end of command character
        DIR_EDGE: '->', indicating a directed edge
        BI_EDGE: '<->', indicating a bidirected edge

It is recommended not to call functions without docstrings in this module, externally.

(CC BY-SA 4.0) 2021. If you use this code, please cite: M.A. Little, R. Badawy,
2019, "Causal bootstrapping", arXiv:1910.09648
"""

import grapl.ply.lex as lex
import grapl.ply.yacc as yacc
import grapl.admg as admg
import copy


class GraplDSL():
    """A GRAPL lexer and parser object, implemented using PLY. See PLY documentation
       in this package, for further details.
    """
    def __init__(self):
        self.lexer = lex.lex(module=self)
        self.parser = yacc.yacc(module=self)

    tokens = (
        'NODE',
        'DIR_EDGE',
        'BI_EDGE',
        'EOC',
        'TITLE'
    )

    t_DIR_EDGE = '->'
    t_BI_EDGE = '<->'
    t_EOC = ';'
    t_ignore = ' \t'

    def t_TITLE(self,t):
        r'\"(.)*\"'
        return t

    def t_NODE(self,t):
        r'([a-zA-Z_])+(\d)*'
        return t

    def t_newline(self,t):
        r'\n+'
        t.lexer.lineno += len(t.value)

    def t_error(self,t):
        raise SyntaxError("Grapl lexical error at line: " + str(t.lineno) + ", near: " + t.value)

    def p_grapl(self,p):
        '''grapl : TITLE EOC graph
                 | graph'''
        if len(p) == 4:
            self.admg.settitle(p[1])

    def p_graph(self,p):
        '''graph : graph command
                | command'''
        pass

    def p_command(self,p):
        '''command : NODE EOC
                | NODE DIR_EDGE NODE EOC
                | NODE BI_EDGE NODE EOC'''
        if len(p) == 3:
            self.admg.addvar(p[1])
        elif p[2] == self.t_DIR_EDGE:
            if p[1] not in self.admg.vars:
                print("Grapl error: node " + p[1] + " not already defined at line: " + str(p.lineno(1)))
            elif p[3] not in self.admg.vars:
                print("Grapl error: node " + p[3] + " not already defined at line: " + str(p.lineno(3)))
            else:
                self.admg.addedges(p[3],parents={p[1]})
        elif p[2] == self.t_BI_EDGE:
            if p[1] not in self.admg.vars:
                print("Grapl error: node " + p[1] + " not already defined at line: " + str(p.lineno(1)))
            elif p[3] not in self.admg.vars:
                print("Grapl error: node " + p[3] + " not already defined at line: " + str(p.lineno(3)))
            else:
                self.admg.addedges(p[3],bidirects={p[1]})
        
    def p_error(self,p):
        if p is not None:
            raise SyntaxError("Grapl syntax error at line: " + str(p.lineno) + ", near: " + p.value)
        else:
            raise SyntaxError("Grapl syntax error: unexpected end of input")

    def readgrapl(self, scm_string=''):
        """Read a GRAPL syntax string describing an ADMG, and create a graph from it. Returns
           an ADMG object; see the description of the ADMG class, for further details. Raises
           a SyntaxError exception if the GRAPL string is not well formed.
        """
        self.admg = admg.ADMG()
        try:
            self.parser.parse(scm_string)
        except SyntaxError as ex:
            print(ex)
            return None
        self.admg.connect()
        return copy.deepcopy(self.admg)
    
    def savegrapl(self, admg):
        """Given an ADMG object as parameter, outputs a corresponding GRAPL syntax string.
           See the description of the ADMG class for further details about this structure.
           Returns a String.
        """
        nodes = set(admg.vars)
        scm_string = admg.title + self.t_EOC + ' '

        for node in nodes:
            scm_string = scm_string + node + self.t_EOC + ' '
        
        for node in nodes:
            parents = admg.pa({node})
            for parent in parents:
                scm_string = scm_string + parent + self.t_DIR_EDGE + node + self.t_EOC + ' '
        
        for node in nodes:
            bidirects = admg.bi({node})
            for bidirect in bidirects:
                scm_string = scm_string + node + self.t_BI_EDGE + bidirect + self.t_EOC + ' '

        return scm_string

    def displaygrapltokens(self, scm_string=''):
        """For debugging, print out the sequence of lexical tokens in a GRAPL syntax string."""
        self.admg = admg.ADMG()
        self.lexer.input(scm_string)
        while True:
            tok = self.lexer.token()
            if not tok: 
                break
            print(tok)



# -*- coding: utf-8 -*-
"""
Unit tests for the GRAPL library, for representing, analyzing and processing
acyclic directed mixed graphs for structural causal modelling.

(CC BY-SA 4.0) 2021. If you use this code, please cite: M.A. Little, R. Badawy,
2019, "Causal bootstrapping", arXiv:1910.09648
"""

import grapl.algorithms as algs
import grapl.dsl as dsl

def run_all():
    test_cyclic()
    test_topsort()
    test_tianfactor()
    test_fixing()
    test_identify1()
    test_identify2()
    test_non_identify1()
    test_non_identify2()
    test_complex_dag()
    test_markov()

    print('All unit tests passed')


def test_cyclic():
    grapl_obj = dsl.GraplDSL()
    grapl_filename = 'grapl/graphs/cyclic_graph.grapl'
    G = grapl_obj.readgrapl(open(grapl_filename, 'r').read())
    assert not G.isdag()
    assert not G.isayclic()

def test_topsort():
    grapl_obj = dsl.GraplDSL()
    grapl_filename = 'grapl/graphs/topsort_graph.grapl'
    G = grapl_obj.readgrapl(open(grapl_filename, 'r').read())
    node_list = G.topsort()
    for node in node_list:
        for parent in G.pa({node}):
            assert node_list.index(node) > node_list.index(parent)

def test_tianfactor():
    grapl_obj = dsl.GraplDSL()
    grapl_filename = 'grapl/graphs/bareinboim_2020.grapl'
    G = grapl_obj.readgrapl(open(grapl_filename, 'r').read())
    assert G.isayclic()
    assert all([(D in G.districts()) for D in [{'A'},{'F'},{'B','D'},{'E','C'}]])
    assert all([(v in G.an({'D'})) for v in {'A','B','C','D'}])
    assert all([(v in G.de({'B','F'})) for v in {'B','D','E','F'}])

def test_fixing():
    grapl_obj = dsl.GraplDSL()
    grapl_filename = 'grapl/graphs/richardson_2017.grapl'
    G = grapl_obj.readgrapl(open(grapl_filename, 'r').read())
    id_expr, expr, isident = algs.idfixing(G, {'X2'}, {'X4'})
    assert not G.isdag()
    assert G.isayclic()
    assert isident

def test_identify1():
    grapl_obj = dsl.GraplDSL()
    grapl_filename = 'grapl/graphs/shpitser_thesis1.grapl'
    G = grapl_obj.readgrapl(open(grapl_filename, 'r').read())
    id_expr, expr, isident = algs.idfixing(G, {'X'}, {'Y1','Y2'})
    assert not G.isdag()
    assert G.isayclic()
    assert isident
    assert all([(term in expr.num) for term in [{'X', "W1'", 'Y1'},{"W1'"},{'Y2'}]])
    assert {'X', "W1'"} in expr.den
    assert "W1'" in expr.mrg

def test_identify2():
    grapl_obj = dsl.GraplDSL()
    grapl_filename = 'grapl/graphs/shpitser_thesis2.grapl'
    G = grapl_obj.readgrapl(open(grapl_filename, 'r').read())
    id_expr, expr, isident = algs.idfixing(G, {'X'}, {'Y'})
    assert not G.isdag()
    assert G.isayclic()
    assert isident

def test_non_identify1():
    grapl_obj = dsl.GraplDSL()
    grapl_filename = 'grapl/graphs/non_identify1.grapl'
    G = grapl_obj.readgrapl(open(grapl_filename, 'r').read())
    id_expr, expr, isident = algs.idfixing(G, {'X'}, {'Y'})
    assert not G.isdag()
    assert G.isayclic()
    assert not isident

def test_non_identify2():
    grapl_obj = dsl.GraplDSL()
    grapl_filename = 'grapl/graphs/non_identify2.grapl'
    G = grapl_obj.readgrapl(open(grapl_filename, 'r').read())
    id_expr, expr, isident = algs.idfixing(G, {'X'}, {'Y'})
    assert not G.isdag()
    assert G.isayclic()
    assert not isident

def test_complex_dag():
    grapl_obj = dsl.GraplDSL()
    grapl_filename = 'grapl/graphs/complex_dag.grapl'
    G = grapl_obj.readgrapl(open(grapl_filename, 'r').read())
    id_str, expr, isdag = algs.truncfactor(G, {'X1'}, {'X3','X4','Y1'})
    assert all([(term in expr.num) for term in [{'Y1','C','X1'},{'X3'},{'X4'},{'C'}]])
    assert {'C','X1'} in expr.den
    assert 'C' in expr.mrg

def test_markov():
    grapl_obj = dsl.GraplDSL()
    grapl_filename = 'grapl/graphs/markov1.grapl'
    G = grapl_obj.readgrapl(open(grapl_filename, 'r').read())
    assert G.nd({'D'}) == {'S','I'}
    assert G.nd({'G'}) == {'D','S','I'}
    assert G.nd({'L'}) == {'D','S','I','G'}
    assert G.nd({'S'}) == {'D','I','G','L'}
    assert G.nd({'I'}) == {'D'}



# -*- coding: utf-8 -*-
"""
Algorithms for representing and manipulating ratios of distribution
functions, used to construct interventional distributions on ADMGs.

(CC BY-SA 4.0) 2021. If you use this code, please cite: M.A. Little, R. Badawy,
2019, "Causal bootstrapping", arXiv:1910.09648
"""

import grapl.util as util


class Expr():
    """A distribution expression object, as a ratio of products of marginalized distributions.
       The member variables are: num (numerator), a List of Sets of Strings, each set being
       a distribution, consisting of a set of random variables (node names); den (denominator)
       same as with num; and mrg (marginals), a Set of Strings, each string being the name of a
       random variable which is marginalized out over the whole ratio distribution.
    """
    def __init__(self,num=[],den=[],mrg=set()):
        self.num = list(num)
        self.den = list(den)
        self.mrg = set(mrg)

    def addvars(self, num=[], den=[], mrg=set()):
        """Adds a new set of numerator, denominator and marginal variables to an expression object.
           The parameters are: num (numerator), a List of Sets of Strings, each set being
           a distribution, consisting of a set of random variables (node names); den (denominator)
           same as with num; and mrg (marginals), a Set of Strings, each string being the name of a
           random variable which is marginalized out.
        """
        if len(num) > 0:
            self.num.append(num)
        if len(den) > 0:
            self.den.append(den)
        if len(mrg) > 0:
            self.mrg.add(mrg)
    
    def subsvar(self, old, new):
        """Substitute the old name for a random variable, with a new name. The parameters are:
           old (old variable) String name for an existing random variable; and new: String
           name to which the old variable will be renamed.
        """
        self.num = [{new if (v == old) else v for v in term} for term in self.num]
        self.den = [{new if (v == old) else v for v in term} for term in self.den]
        self.mrg = {new if (v == old) else v for v in self.mrg}
    
    def fix(self, fix, blanket):
        """Apply the interventional fixing operation to a distribution expression. Parameters
           are fix: a String naming the random variable to fix; and blanket: a Set of Strings,
           giving the effective parents of the variable in fix.
        """
        if len(blanket) == 0:
            self.addvars(den={fix})
        else:
            self.addvars(num=blanket, den=blanket.union({fix}))

    def fixmarginal(self, fix):
        """Apply the interventional fixing operation where it can be marginalized. The fix
           parameter is a string giving the name of the random variable to fix. To avoid
           naming clashes, a prime character is appended to the fix variable, and any
           random variables with the name in fix, are renamed accordingly.
        """
        new_var = fix + chr(39)     # Avoid variable name clashes
        self.addvars(mrg = new_var)
        self.subsvar(fix, new_var)

    def cancel(self):
        """An algorithm for cancelling variables in a distribution expression. This seeks to greedily
           match and remove terms appearing in both numerator and denominator of an expression.
           Returns True if any changes to the expression occurred as a result, and False
           otherwise.
        """
        num = self.num.copy()
        den = self.den.copy()
        changes = False
        for term in self.num:
            if term in den:
                den.remove(term)
                num.remove(term)
                changes = True
        if changes:
            self.num = num.copy()
            self.den = den.copy()
        return changes

    def marginal(self):
        """An algorithm for marginalizing out variables in a distribution expression. Greedily
           removes variables appearing in both the numerator and the set of marginal variables.
           Returns True if any changes to the expression occurred as a result, and False
           otherwise.
        """
        num = self.num.copy()
        mrg = self.mrg.copy()
        changes = False
        for node in self.mrg:
            num_cnt = sum([sum({1 if (v == node) else 0 for v in term}) for term in self.num])
            den_find = any([any({(v == node) for v in term}) for term in self.den])
            if (num_cnt == 1) and not den_find:
                changes = True
                num = [term.difference({node}) for term in num]
                mrg.remove(node)
        if changes:
            self.num = num.copy()
            self.mrg = mrg.copy()
        return changes

    def simplify(self):
        """An algorithm for simplifying a distribution, by successive cancellation and marginalization
           until a fixed point is reached. Returns True if simplifications were possible, and
           False otherwise."""
        changes = True
        simplified = False
        while changes:
            change1 = self.cancel()
            change2 = self.marginal()
            changes = change1 or change2
            simplified = simplified or changes
        return simplified

    def tostr(self):
        """Convert a distribution expression to a Latex expression for display. Returns a Latex
           format String."""
        expr_str = ''
        if len(self.mrg) > 0:
            expr_str = expr_str + util.mrgstr(self.mrg)
        if len(self.num) > 1:
            expr_str = expr_str + '['
        for term in self.num:
            expr_str = expr_str + util.probstr(term)
        if len(self.den) > 0:
            expr_str = expr_str + '/'
            if len(self.den) > 1:
                expr_str = expr_str + '{'
            for term in self.den:
                expr_str = expr_str + util.probstr(term)
            if len(self.den) > 1:
                expr_str = expr_str + '}'
        if (len(self.num) > 1):
            expr_str = expr_str + ']'
        return expr_str

    def tocondstr(self):
        """Convert a distribution expression to conditional form, and then into Latex format. This
           greedily matches terms in the numerator with corresponding terms in the denominator,
           so as to represent ratios of terms as conditionals in the numerator. Returns a Latex
           format String.
        """
        expr_str = ''
        if len(self.mrg) > 0:
            expr_str = expr_str + util.mrgstr(self.mrg)
        if len(self.num) > 1:
            expr_str = expr_str + '['
        num_ord = sorted(self.num,key=len,reverse=True)
        den_ord = sorted(self.den,key=len,reverse=True)
        while len(num_ord) > 0:
            next_num = num_ord[0]
            num_ord = num_ord[1:]
            assign_term = False
            den_list = den_ord.copy()
            while (len(den_list) > 0) and (not assign_term):
                next_den = den_list[0]
                den_list = den_list[1:]
                if next_den.issubset(next_num):
                    den_ord.remove(next_den)
                    assign_term = True
                    expr_str = expr_str + util.probcndstr(next_num.difference(next_den), next_den)
            if not assign_term:
                expr_str = expr_str + util.probstr(next_num)
        if len(den_ord) > 0:
            expr_str = expr_str + '/'
            for den in den_ord:
                expr_str = expr_str + util.probstr(den)
        if (len(self.num) > 1):
            expr_str = expr_str + ']'
        return expr_str



# -*- coding: utf-8 -*-
"""
Various utility functions for the GRAPL library.

(CC BY-SA 4.0) 2021. If you use this code, please cite: M.A. Little, R. Badawy,
2019, "Causal bootstrapping", arXiv:1910.09648
"""

import grapl.admg as admg

def csepstr(nodes=[]):
    """Create a comma-separated string for a given list of nodes, given as a List of Strings,
       each string being the name of a random variable. If empty, then an empty string is
       returned. Returns a String.
    """
    return ','.join(str(s) for s in nodes)

def mrgstr(nodes):
    """Create a Latex marginalization string from a list of nodes, provided as a List of Strings,
       each string being the name of a random variable. If empty, then an empty string is
       returned. Returns a String."""
    if len(nodes) > 0:
        return '\sum_{' + csepstr(nodes) + '}'
    else:
        return ''

def phistr(nodes):
    """Create a Latex fixing operation string from a list of nodes, provided as a List of Strings,
       each string being the name of a random variable. If empty, then an empty string is
       returned. Returns a String.
    """
    if len(nodes) > 0:
        return '\phi_{' + csepstr(nodes) + '}'
    else:
        return ''

def probstr(nodes):
    """Create a Latex distribution string from a list of nodes, given as a List of Strings,
       each string being the name of a random variable. If empty, then an empty string is
       returned. Returns a String.
    """
    if len(nodes) > 0:
        return 'p(' + csepstr(nodes) + ')'
    else:
        return ''

def probcndstr(nodes=[], cnd_nodes=[]):
    """Create a Latex conditional distribution string from a list of nodes and conditional nodes.
       The parameters are Lists of Strings; nodes being the list of conditioned random variables,
       and cnd_nodes being the list of their parents; if the latter is empty then the distribution
       string is not conditional, and if the former is empty then an empty string is returned.
       Returns a String.
    """
    prob_str = ''
    if len(nodes) > 0:
        prob_str = prob_str + 'p(' + csepstr(nodes)
        if len(cnd_nodes) > 0:
            prob_str = prob_str + '|' + csepstr(cnd_nodes)
        prob_str = prob_str + ')'
    return prob_str



# -*- coding: utf-8 -*-
"""
Tutorial illustrating the use of the GRAPL library, for representing, analyzing and processing
acyclic directed mixed graphs for structural causal modelling.

(CC BY-SA 4.0) 2021. If you use this code, please cite: M.A. Little, R. Badawy,
2019, "Causal bootstrapping", arXiv:1910.09648
"""

import grapl.algorithms as algs
import grapl.dsl as dsl

# Create a GRAPL DSL parser
grapl_obj = dsl.GraplDSL()

# Create a DAG from a GRAPL string
dag_grapl = ' "Symptoms"; \
    Sinus; Headache; Nose; Flu; Allergy; \
    Sinus -> Nose; \
    Flu -> Sinus; \
    Sinus -> Headache; \
    Allergy -> Sinus;'
G = grapl_obj.readgrapl(dag_grapl)
G.display()

print('\nCheck if the DAG acyclic:')
print(G.isdag()) # True

print('\nTopological sort of DAG nodes:')
print(G.topsort()) # ['Flu', 'Allergy', 'Sinus', 'Nose', 'Headache']

print('\nAll ancestors of Headache:')
print(G.an({'Headache'})) # {'Allergy', 'Headache', 'Sinus', 'Flu'}

print('\nAll DAG local Markov conditional independence relations:')
ci, isdag = algs.localmarkov(G) # {'Flu⊥Allergy', 'Nose⊥Headache,Allergy,Flu|Sinus', 'Allergy⊥Flu', 'Headache⊥Nose,Allergy,Flu|Sinus'}
print(ci)

print('\nFactorized joint distribution over Headache,Allergy,Flu:')
factor_str, expr, isdag = algs.dagfactor(G, {'Headache','Allergy','Flu'})
print(factor_str) # p(Flu,Allergy,Headache)=\sum_{Sinus}[p(Sinus|Flu,Allergy)p(Headache|Sinus)p(Allergy)p(Flu)]

print('\nInterventional (cause-effect) distribution of Sinus on Headache,Allergy,Flu:')
id_str, expr, isdag = algs.truncfactor(G, {'Sinus'}, {'Headache','Allergy','Flu'})
print(id_str) # p_{Sinus}(Flu,Allergy,Headache)=[p(Headache|Sinus)p(Allergy)p(Flu)]



# -*- coding: utf-8 -*-
"""
Implements algorithms for processing of structural causal models represented
by acyclic directed mixed graphs (ADMGs).

(CC BY-SA 4.0) 2021. If you use this code, please cite: M.A. Little, R. Badawy,
2019, "Causal bootstrapping", arXiv:1910.09648
"""

import grapl.admg as admg
import grapl.expr as expr
import grapl.util as util
import copy as cp


def idfixing(G, X={}, Y={}):
    """Implementation of Richardson et al.'s Theorem 60 for causal effect identification
       from given ADMG. See: T.S. Richardson, J.M. Robins, I. Shpitser, 2012:
       "Nested Markov properties for acyclic directed mixed graphs", UAI'12.

       Parameters:
       G (ADMG)           -- A mixed causal graph object
       X (Set of Strings) -- Set of interventional variables, where each string is a random variable name (must not be empty)
       Y (Set of Strings) -- Set of variables on which interventional expression is represented, where each string is a random
                             variable name (if empty, then all variables in G except for those in X)

       Returns: (String, Expr, Boolean)
       If interventional distribution is identifiable, returns the identified expression string
       the corresponding identified Expr object, and True. Otherwise, returns '', None, False
    """

    # Must supply the set of interventional (cause) variables
    if not X:
        return '', None, False

    # Collate variable names
    V = G.nodes()

    # Handle effects
    if not Y:
        Y = V.difference(X)

    # Find Y* = an_G(V)_{V\A}(Y)
    A = X
    VdA = V.difference(A)
    G_sub_VdA = G.sub(VdA)
    Y_star = G_sub_VdA.an(Y)

    # Find subgraph G(V)_Y* and its districts
    G_sub_Ys = G.sub(Y_star)
    D_y = G_sub_Ys.districts()

    # For each district, determine post-fixed ADMG and corresponding expression
    D_exprs = []
    for D in D_y:

        # Nodes to fix for this district, fixing transform: phi_{V\D}(p(V;G_V))
        VdD = V.difference(D)

        # Initialize accumulated ADMG, G(V), and district expression, p(V)
        G_fixed = cp.deepcopy(G)
        D_expr = expr.Expr(num=[V])

        # Try to apply fixing sequence
        nodes_fix = VdD.copy()
        while (len(nodes_fix) > 0):
            
            # Find a valid fixing node in sequence
            fixable_nodes = {node for node in nodes_fix if G_fixed.isfixable(node)}
            node_valid = (len(fixable_nodes) > 0)
            if node_valid:
                fix = fixable_nodes.pop()
            else:
                break
            
            # Fix or marginalize (Corollary 33) the accumulated expression
            dispa_fix = G_fixed.dispa(fix)
            fix_ch = G_fixed.ch({fix})
            if len(fix_ch) == 0:
                D_expr.fixmarginal(fix)
            else:
                D_expr.fix(fix, dispa_fix)

            # Fix node (remove all its incoming edges) in accumulated ADMG,
            # remove from remaining fixing sequence
            G_fixed.fix(fix)
            nodes_fix = nodes_fix.difference({fix})

        if not node_valid:
            break
        
        D_expr.simplify()
        D_exprs.append(D_expr)

    # At this point, if no valid fixing node, intervention is not identifiable
    if not node_valid:
        return '', None, False
    else:
        Y_mrg = Y_star.difference(Y)
        id_expr = expr.Expr(mrg=Y_mrg)
        for ex in D_exprs:
            clash_vars = id_expr.mrg.difference(id_expr.mrg.difference(ex.mrg))
            for old_var in clash_vars:
                new_var = old_var + chr(39)
                ex.subsvar(old_var, new_var)
            id_expr.num = id_expr.num + ex.num
            id_expr.den = id_expr.den + ex.den
            id_expr.mrg = id_expr.mrg.union(ex.mrg)

        id_expr.simplify()
        id_str = 'p_{'+util.csepstr(A)+'}('+util.csepstr(Y)+')='
        id_str = id_str+id_expr.tocondstr()

        return id_str, id_expr, True


def dagfactor(G, Y={}, simplify=True):
    """Factorized distribution for DAGs (ADMGs with no bidirects).

       Parameters:
       G (ADMG)           -- DAG representing the causal graph (must not have bidirects)
       Y (Set of Strings) -- Set of effect variables (if empty, all variables in G)
       simplify (Boolean) -- If True, final expression is explicitly simplified

       Returns: (String, Expr, Boolean)
       If G is a DAG, factored expression string, corresponding Expr object, and True.
       If G is not a DAG then '', None, False.
    """

    # Check to make sure there are no latent variables
    if not G.isdag():
        return '', None, False

    # Collate variable names
    V = G.nodes()

    # Handle marginals
    if not Y:
        Y = V
    Y_mrg = V.difference(Y)
    fac_expr = expr.Expr(mrg=Y_mrg)

    # Construct factorized expression over all nodes
    for node in V:
        fac_expr.num = fac_expr.num + [G.pa({node}).union({node})]
        fac_expr.den = fac_expr.den + [G.pa({node})]

    # Final expression simplification
    if simplify:
        fac_expr.simplify()

    # Latex expression construction
    dist_str = 'p('+util.csepstr(Y)+')='
    dist_str = dist_str + fac_expr.tocondstr()

    return dist_str, fac_expr, True
    

def truncfactor(G, X={}, Y={}, prefactor=True):
    """Truncated factorization ("g-formula") for DAGs (ADMGs with no bidirects).

       Parameters:
       G (ADMG) -- DAG object representing the causal graph (must not have bidirects)
       X (Set of Strings)  -- Interventional variables where each string is a random variable name (must not be empty)
       Y (Set of Strings)  -- Effect variables, each sring is a random variable name (if empty, all variables in G other than the set X)
       prefactor (Boolean) -- If True, joint distribution is chain factored before fixing

       Returns: (String, Expr, Boolean)
       If G is a DAG, factored interventonal distribution string, corresponding Expr object,
       and True. Otherwise, returns '', None, False.
    """

    # Must supply the set of interventional variables
    if not X:
        return '', None, False

    # Check there are no latent variables
    if not G.isdag():
        return '', None, False

    # Collate variable names
    V = G.nodes()

    # Initialize accumulated DAG and interventional expression, p(V)
    G_fixed = cp.deepcopy(G)
    if prefactor:
        dummy, D_expr, dummy_bool = dagfactor(G)        # Start with factorized joint distribution
    else:
        D_expr = expr.Expr(num=[V])                     # Start with the unfactored joint distribution

    # Apply fixing sequence
    nodes_fix = cp.deepcopy(X)
    while (len(nodes_fix) > 0):
        
        # Select next fixing node in sequence
        fix = nodes_fix.pop()
        
        # Fix or marginalize (Corollary 33) the accumulated expression
        dispa_fix = G_fixed.dispa(fix)
        fix_ch = G_fixed.ch({fix})
        if len(fix_ch) == 0:
            D_expr.fixmarginal(fix)
        else:
            D_expr.fix(fix, dispa_fix)

        # Fix node (remove all its incoming edges) in accumulated DAG,
        # remove from remaining fixing sequence
        G_fixed.fix(fix)
        nodes_fix = nodes_fix.difference({fix})

    # Simplify final expression
    D_expr.simplify()

    # Handle marginals
    if not Y:
        Y = V.difference(X)
    Y_mrg = (V.difference(Y)).difference(X)
    id_expr = expr.Expr(mrg=Y_mrg)

    # Avoid naming clashes with dummy marginal variables
    clash_vars = id_expr.mrg.difference(id_expr.mrg.difference(D_expr.mrg))
    for old_var in clash_vars:
        new_var = old_var + chr(39)
        D_expr.subsvar(old_var, new_var)
    id_expr.num = id_expr.num + D_expr.num
    id_expr.den = id_expr.den + D_expr.den
    id_expr.mrg = id_expr.mrg.union(D_expr.mrg)

    # Final simplification and Latex expression construction
    id_expr.simplify()
    id_str = 'p_{'+util.csepstr(X)+'}('+util.csepstr(Y)+')='
    id_str = id_str + id_expr.tocondstr()

    return id_str, id_expr, True


def localmarkov(G):
    """Compute all local Markov independences for DAGs (ADMGs with no bidirects).

       Parameters:
       G (ADMG) -- DAG object representing the causal graph (must not have bidirects)

       Returns: (Set of Strings, Boolean)
       If G is a DAG, set of strings representing Markov independences, True. Otherwise returns empty set, False.
    """

    # Check to make sure there are no latent variables
    if not G.isdag():
        return set(), False

    # Collate variable names
    nodes = set(G.vars)

    ci = set()
    for node in nodes:
        parents = G.pa({node})
        # parents = G.pa({node}).difference({node})
        non_descend = G.nd({node})
        conds = non_descend.difference({node}).difference(parents)
        if conds:
            li = node + '\u22a5' + util.csepstr(conds)
            if parents:
                li = li + '|' + util.csepstr(parents)
            ci.add(li)

    return ci, True


def admgfactor(G, Y={}):
    """Factorized distribution for ADMGs.

       Parameters:
       G (ADMG)           -- A mixed causal graph
       Y (Set of Strings) -- Set of variables on which expression is represented, where each string is a random variable name
                             (if empty, then all variables in G)

       Returns: (String, Expr)
       Factored distribution expression string, corresponding Expr object.
    """

    # Collate variable names
    V = G.nodes()

    # Find a topological ordering of the observed nodes
    sort_nodes = G.topsort()

    # Handle marginals
    if not Y:
        Y = V
    Y_mrg = V.difference(Y)
    fac_expr = expr.Expr(mrg=Y_mrg)

    # Factor distribution for each observed node in turn
    for node in V:

        # Set of all nodes preceding and including this one in topological order
        nodes_prec = set(sort_nodes[:sort_nodes.index(node) + 1])

        # Subgraph of all such ordered nodes
        G_prec = G.sub(nodes_prec)

        # District for node in G_prec
        D_node = G_prec.district(node)

        # Restrict ordered nodes to those within the district
        D_node_prec = nodes_prec.intersection(D_node)

        # Find (inclusive) parents of all such nodes
        D_parents = G.pa(D_node_prec).union(D_node_prec)

        # Remove node to find 'shielding' nodes
        shield = D_parents.difference({node})

        fac_expr.num = fac_expr.num + [shield.union({node})]
        fac_expr.den = fac_expr.den + [shield]

    # Final expression simplification
    fac_expr.simplify()

    # Construct Latex expression
    dist_str = 'p('+util.csepstr(Y)+')='
    dist_str = dist_str + fac_expr.tocondstr()

    return dist_str, fac_expr




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'BI_EDGE DIR_EDGE EOC NODE TITLEgrapl : TITLE EOC graph\n                 | graphgraph : graph command\n                | commandcommand : NODE EOC\n                | NODE DIR_EDGE NODE EOC\n                | NODE BI_EDGE NODE EOC'
    
_lr_action_items = {'TITLE':([0,],[2,]),'NODE':([0,3,4,6,7,8,9,10,11,14,15,],[5,5,-4,5,-3,-5,12,13,5,-6,-7,]),'$end':([1,3,4,7,8,11,14,15,],[0,-2,-4,-3,-5,-1,-6,-7,]),'EOC':([2,5,12,13,],[6,8,14,15,]),'DIR_EDGE':([5,],[9,]),'BI_EDGE':([5,],[10,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'grapl':([0,],[1,]),'graph':([0,6,],[3,11,]),'command':([0,3,6,11,],[4,7,4,7,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> grapl","S'",1,None,None,None),
  ('grapl -> TITLE EOC graph','grapl',3,'p_grapl','dsl.py',66),
  ('grapl -> graph','grapl',1,'p_grapl','dsl.py',67),
  ('graph -> graph command','graph',2,'p_graph','dsl.py',72),
  ('graph -> command','graph',1,'p_graph','dsl.py',73),
  ('command -> NODE EOC','command',2,'p_command','dsl.py',77),
  ('command -> NODE DIR_EDGE NODE EOC','command',4,'p_command','dsl.py',78),
  ('command -> NODE BI_EDGE NODE EOC','command',4,'p_command','dsl.py',79),
]



from setuptools import setup

setup(
    name='grapl-causal',
    version='1.4',
    description='GRAPL: A computational library for nonparametric structural causal modelling, analysis and inference',
    url='https://github.com/max-little/GRAPL',
    author='Max A. Little',
    author_email='maxl@mit.edu',
    license='GNU GPL-3.0',
    packages=['grapl','grapl.test','grapl.ply','grapl.tutorials'],
    package_data={'grapl': ['graphs/*.grapl']},
    install_requires=[],
)
